from __future__ import annotations

from typing import TYPE_CHECKING

from soda.execution.query import Query
from soda.profiling.profile_columns_result import ProfileColumnsResult
from soda.profiling.profile_columns_result_column import ProfileColumnsResultColumn
from soda.profiling.profile_columns_result_table import ProfileColumnsResultTable
from soda.sodacl.profile_columns_cfg import ProfileColumnsCfg

if TYPE_CHECKING:
    from soda.execution.data_source_scan import DataSourceScan


class ProfileColumnsRun:
    def __init__(self, data_source_scan: DataSourceScan, profile_columns_cfg: ProfileColumnsCfg):

        self.data_source_scan = data_source_scan
        self.soda_cloud = data_source_scan.scan._configuration.soda_cloud
        self.data_source = data_source_scan.data_source
        self.profile_columns_cfg: ProfileColumnsCfg = profile_columns_cfg
        self.logs = self.data_source_scan.scan._logs

    def run(self) -> ProfileColumnsResult:
        profile_columns_result: ProfileColumnsResult = ProfileColumnsResult(self.profile_columns_cfg)
        self.logs.info(f"Running column profiling for data source: {self.data_source.data_source_name}")

        # row_counts is a dict that maps table names to row counts.
        row_counts_by_table_name: dict[str, int] = self.data_source.get_row_counts_all_tables(
            include_tables=self._get_table_expression(self.profile_columns_cfg.include_columns),
            exclude_tables=self._get_table_expression(self.profile_columns_cfg.exclude_columns),
            query_name="profile columns: get tables and row counts",
        )
        parsed_tables_and_columns = self._build_column_inclusion(self.profile_columns_cfg.include_columns)
        for table_name in row_counts_by_table_name:
            self.logs.debug(f"Profiling columns for {table_name}")
            measured_row_count = row_counts_by_table_name[table_name]
            profile_columns_result_table = profile_columns_result.create_table(
                table_name, self.data_source.data_source_name, measured_row_count
            )

            # get columns & metadata for current table
            columns_metadata_sql = self.data_source.sql_to_get_column_metadata_for_table(table_name)
            columns_metadata_query = Query(
                data_source_scan=self.data_source_scan,
                unqualified_query_name=f"get col metadata for table: {table_name}",
                sql=columns_metadata_sql,
            )
            columns_metadata_query.execute()
            assert columns_metadata_query.rows, f"No metadata was captured for table: {table_name}"
            columns_metadata_result = {column[0]: column[1] for column in columns_metadata_query.rows}

            # perform numerical metrics collection
            numerical_columns = {
                col_name: data_type
                for col_name, data_type in columns_metadata_result.items()
                if data_type in self.data_source.NUMERIC_TYPES_FOR_PROFILING
            }

            for column_name, column_type in numerical_columns.items():
                self.profile_numeric_column(
                    column_name,
                    column_type,
                    table_name,
                    columns_metadata_result,
                    parsed_tables_and_columns,
                    profile_columns_result_table,
                )

            # text columns
            text_columns = {
                col_name: data_type
                for col_name, data_type in columns_metadata_result.items()
                if data_type in self.data_source.TEXT_TYPES_FOR_PROFILING
            }
            for column_name, column_type in text_columns.items():
                self.profile_text_column(
                    column_name,
                    column_type,
                    table_name,
                    columns_metadata_result,
                    parsed_tables_and_columns,
                    profile_columns_result_table,
                )

        if not profile_columns_result.tables:
            self.logs.error(f"Profiling for data source: {self.data_source.data_source_name} failed")
        return profile_columns_result

    def profile_numeric_column(
        self,
        column_name: str,
        column_type: str,
        table_name: str,
        columns_metadata_result: dict,
        parsed_tables_and_columns: dict[str, list[str]],
        profile_columns_result_table: ProfileColumnsResultTable,
    ):
        self.logs.debug(f"Profiling column {column_name} of {table_name}")
        profile_columns_result_column, is_included_column = self.build_profiling_column(
            column_name,
            column_type,
            table_name,
            list(columns_metadata_result.keys()),
            parsed_tables_and_columns,
            profile_columns_result_table,
        )
        if profile_columns_result_column and is_included_column:
            value_frequencies_sql = self.data_source.profiling_sql_values_frequencies_query(table_name, column_name)

            value_frequencies_query = Query(
                data_source_scan=self.data_source_scan,
                unqualified_query_name=f"profiling: {table_name}, {column_name}: mins, maxes and values frequencies",
                sql=value_frequencies_sql,
            )
            value_frequencies_query.execute()
            if value_frequencies_query.rows is not None:
                profile_columns_result_column.mins = [
                    float(row[0]) if not isinstance(row[0], int) else row[0] for row in value_frequencies_query.rows
                ]
                profile_columns_result_column.maxes = [
                    float(row[1]) if not isinstance(row[1], int) else row[1] for row in value_frequencies_query.rows
                ]
                profile_columns_result_column.min = profile_columns_result_column.mins[0]
                profile_columns_result_column.max = profile_columns_result_column.maxes[0]
                profile_columns_result_column.frequent_values = self.build_frequent_values_dict(
                    values=[row[2] for row in value_frequencies_query.rows],
                    frequencies=[row[3] for row in value_frequencies_query.rows],
                )
            else:
                self.logs.error(
                    f"Database returned no results for minumum values, maximum values and frequent values in table: {table_name}, columns: {column_name}"
                )

            # pure aggregates
            aggregates_sql = self.data_source.profiling_sql_numeric_aggregates(table_name, column_name)
            aggregates_query = Query(
                data_source_scan=self.data_source_scan,
                unqualified_query_name=f"profiling: {table_name}, {column_name}: get_pure_profiling_aggregates",
                sql=aggregates_sql,
            )
            aggregates_query.execute()
            if aggregates_query.rows is not None:
                profile_columns_result_column.average = float(aggregates_query.rows[0][0])
                profile_columns_result_column.sum = float(aggregates_query.rows[0][1])
                profile_columns_result_column.variance = float(aggregates_query.rows[0][2])
                profile_columns_result_column.standard_deviation = float(aggregates_query.rows[0][3])
                profile_columns_result_column.distinct_values = int(aggregates_query.rows[0][4])
                profile_columns_result_column.missing_values = int(aggregates_query.rows[0][5])
            else:
                self.logs.error(
                    f"Database returned no results for aggregates in table: {table_name}, columns: {column_name}"
                )

            # histogram
            assert (
                profile_columns_result_column.min is not None
            ), "Min cannot be None, make sure the min metric is derived before histograms"
            assert (
                profile_columns_result_column.max is not None
            ), "Max cannot be None, make sure the min metric is derived before histograms"
            histogram_sql, bins_list = self.data_source.histogram_sql_and_boundaries(
                table_name, column_name, profile_columns_result_column.min, profile_columns_result_column.max
            )
            histogram_query = Query(
                data_source_scan=self.data_source_scan,
                unqualified_query_name=f"profiling: {table_name}, {column_name}: get histogram",
                sql=histogram_sql,
            )
            histogram_query.execute()
            histogram = {}
            if histogram_query.rows is not None:
                histogram["boundaries"] = bins_list
                histogram["frequencies"] = [int(freq) if freq is not None else 0 for freq in histogram_query.rows[0]]
                profile_columns_result_column.histogram = histogram
            else:
                self.logs.error(
                    f"Database returned no results for histograms in table: {table_name}, columns: {column_name}"
                )
        else:
            self.logs.info(f"No profiling information derived for column {column_name} in {table_name}")

    def profile_text_column(
        self,
        column_name: str,
        column_type: str,
        table_name: str,
        columns_metadata_result: dict,
        parsed_tables_and_columns: dict[str, list[str]],
        profile_columns_result_table: ProfileColumnsResultTable,
    ):
        profile_columns_result_column, is_included_column = self.build_profiling_column(
            column_name,
            column_type,
            table_name,
            list(columns_metadata_result.keys()),
            parsed_tables_and_columns,
            profile_columns_result_table,
        )
        if profile_columns_result_column and is_included_column:
            # frequent values for text column
            frequent_values_sql = self.data_source.profiling_sql_top_values(table_name, column_name)
            frequent_values_query = Query(
                data_source_scan=self.data_source_scan,
                unqualified_query_name=f"profiling: {table_name}, {column_name}: get frequent values text cols",
                sql=frequent_values_sql,
            )
            frequent_values_query.execute()
            if frequent_values_query.rows:
                profile_columns_result_column.frequent_values = self.build_frequent_values_dict(
                    values=[row[2] for row in frequent_values_query.rows],
                    frequencies=[row[0] for row in frequent_values_query.rows],
                )
            else:
                self.logs.error(
                    f"Database returned no results for textual frequent values in {table_name}, column: {column_name}"
                )
            # pure text aggregates
            text_aggregates_sql = self.data_source.profiling_sql_text_aggregates(table_name, column_name)
            text_aggregates_query = Query(
                data_source_scan=self.data_source_scan,
                unqualified_query_name=f"profiling: {table_name}, {column_name}: get textual aggregates",
                sql=text_aggregates_sql,
            )
            text_aggregates_query.execute()
            if text_aggregates_query.rows:
                profile_columns_result_column.distinct_values = int(text_aggregates_query.rows[0][0])
                profile_columns_result_column.missing_values = int(text_aggregates_query.rows[0][1])
                profile_columns_result_column.average_length = (
                    int(text_aggregates_query.rows[0][2]) if text_aggregates_query.rows[0][2] is not None else None
                )
                profile_columns_result_column.min_length = (
                    int(text_aggregates_query.rows[0][3]) if text_aggregates_query.rows[0][3] is not None else None
                )
                profile_columns_result_column.max_length = (
                    int(text_aggregates_query.rows[0][4]) if text_aggregates_query.rows[0][4] is not None else None
                )
            else:
                self.logs.error(
                    f"Database returned no results for textual aggregates in table: {table_name}, columns: {column_name}"
                )
        else:
            self.logs.info(f"No profiling information derived for column {column_name} in {table_name}")

    def build_profiling_column(
        self,
        column_name: str,
        column_type: str,
        table_name: str,
        table_columns: list[str],
        parsed_tables_and_columns: dict[str, list[str]],
        table_result: ProfileColumnsResultTable,
    ) -> tuple[ProfileColumnsResultColumn | None, bool]:
        column_name = column_name.lower()
        if self._is_column_included_for_profiling(column_name, table_name, table_columns, parsed_tables_and_columns):
            profile_columns_result_column: ProfileColumnsResultColumn = table_result.create_column(
                column_name, column_type
            )
            return profile_columns_result_column, True
        return None, False

    @staticmethod
    def build_frequent_values_dict(values: list[str | int | float], frequencies: list[int]) -> list[dict[str, int]]:
        frequent_values = []
        for i, value in enumerate(values):
            frequent_values.append({"value": str(value), "frequency": frequencies[i]})
        return frequent_values

    def _is_column_included_for_profiling(
        self,
        candidate_column_name: str,
        table_name: str,
        table_columns: list[str],
        parsed_tables_and_columns: dict[str, list[str]],
    ):

        table_name = table_name.lower()
        candidate_column_name = candidate_column_name.lower()
        cols_for_all_tables = parsed_tables_and_columns.get("%", [])
        table_columns = [col_name.lower() for col_name in table_columns]
        if (
            candidate_column_name in parsed_tables_and_columns.get(table_name, [])
            or candidate_column_name in cols_for_all_tables
            or parsed_tables_and_columns.get(table_name, []) == ["%"]
        ):
            if candidate_column_name in table_columns:
                return True
        else:
            return False

    # TODO: Deal with exclude set as well
    def _build_column_inclusion(self, columns_expression: list[str]) -> dict[str, list[str]]:
        included_columns = {}
        for col_expression in columns_expression:
            table, column = col_expression.split(".")
            table = table.lower()
            if table in included_columns.keys():
                if included_columns[table] is None:
                    included_columns[table] = [column]
                else:
                    included_columns[table].append(column)
            else:
                included_columns.update({table: [column]})
        return included_columns

    def _get_table_expression(self, columns_expression: list[str]) -> list[str]:
        table_expressions = []
        for column_expression in columns_expression:
            parts = column_expression.split(".")
            if len(parts) != 2:
                self.logs.error(
                    f'Invalid include column expression "{column_expression}"',
                    location=self.profile_columns_cfg.location,
                )
            else:
                table_expression = parts[0]
                table_expressions.append(table_expression)
        return table_expressions

    def _get_colums(self, columns_expression: list[str]) -> list[str]:
        column_expressions = []
        for column_expression in columns_expression:
            parts = column_expression.split(".")
            if len(parts) != 2:
                self.logs.error(
                    f"Invalid include columns expression '{column_expression}'",
                    location=self.profile_columns_cfg.location,
                )
            else:
                column_expression = parts[1]
                column_expressions.append(column_expression)
        return column_expressions



from numbers import Number
from typing import Dict, Optional

from soda.execution.check import Check
from soda.execution.check_outcome import CheckOutcome
from soda.execution.derived_metric import DERIVED_METRIC_NAMES
from soda.execution.metric import Metric
from soda.execution.user_defined_numeric_metric import UserDefinedNumericMetric
from soda.sodacl.metric_check_cfg import MetricCheckCfg

KEY_CHECK_VALUE = "check_value"


class MetricCheck(Check):
    def __init__(
        self,
        check_cfg: "MetricCheckCfg",
        data_source_scan: "DataSourceScan",
        partition: Optional["Partition"] = None,
        column: Optional["Column"] = None,
    ):
        metric_check_cfg: MetricCheckCfg = check_cfg
        metric_name = metric_check_cfg.metric_name

        super().__init__(
            check_cfg=check_cfg, data_source_scan=data_source_scan, partition=partition, column=column, name=metric_name
        )

        self.historic_descriptor = None
        self.metric_store_request = None

        self.check_value: Optional[float] = None
        self.formula_values: Optional[dict] = None
        self.historic_diff_values: Optional[dict] = None

        from soda.execution.derived_metric import DerivedMetric
        from soda.execution.numeric_query_metric import NumericQueryMetric

        is_user_defined_metric_query = metric_check_cfg.metric_query is not None
        if not is_user_defined_metric_query:
            is_derived = metric_name in DERIVED_METRIC_NAMES

            if not is_derived:
                metric = NumericQueryMetric(
                    data_source_scan=self.data_source_scan,
                    partition=self.partition,
                    column=self.column,
                    metric_name=metric_name,
                    metric_args=metric_check_cfg.metric_args,
                    filter=metric_check_cfg.filter,
                    aggregation=metric_check_cfg.metric_expression,
                    check_missing_and_valid_cfg=metric_check_cfg.missing_and_valid_cfg,
                    column_configurations_cfg=self.column.column_configurations_cfg if self.column else None,
                    check=self,
                )
            else:
                metric = DerivedMetric(
                    data_source_scan=self.data_source_scan,
                    partition=self.partition,
                    column=self.column,
                    metric_name=metric_name,
                    metric_args=metric_check_cfg.metric_args,
                    filter=metric_check_cfg.filter,
                    check_missing_and_valid_cfg=metric_check_cfg.missing_and_valid_cfg,
                    column_configurations_cfg=self.column.column_configurations_cfg if self.column else None,
                    check=self,
                )

            metric = data_source_scan.resolve_metric(metric)

        else:
            if data_source_scan.data_source.is_supported_metric_name(metric_name):
                self.logs.warning(
                    f"User defined metric {metric_name} is a data_source supported metric. Please, choose a different name for the metric.",
                    location=metric_check_cfg.location,
                )

            metric = UserDefinedNumericMetric(
                data_source_scan=self.data_source_scan,
                check_name=metric_check_cfg.source_line,
                sql=metric_check_cfg.metric_query,
                check=self,
            )

            metric = data_source_scan.resolve_metric(metric)

        self.metrics[metric_name] = metric

    def evaluate(self, metrics: Dict[str, Metric], historic_values: Dict[str, object]):
        self.check_value = self.get_metric_value()
        self.set_outcome_based_on_check_value()

    def get_metric_value(self) -> object:
        metric = self.get_metric()
        self.formula_values = metric.formula_values
        self.failed_rows_sample_ref = metric.failed_rows_sample_ref
        return metric.value

    def get_metric(self):
        metric_check_cfg: MetricCheckCfg = self.check_cfg
        metric_name = metric_check_cfg.metric_name
        metric = self.metrics.get(metric_name)
        return metric

    def set_outcome_based_on_check_value(self):
        metric_check_cfg: MetricCheckCfg = self.check_cfg
        if self.check_value is not None and metric_check_cfg.has_threshold():
            if isinstance(self.check_value, Number):
                if metric_check_cfg.fail_threshold_cfg and metric_check_cfg.fail_threshold_cfg.is_bad(self.check_value):
                    self.outcome = CheckOutcome.FAIL
                elif metric_check_cfg.warn_threshold_cfg and metric_check_cfg.warn_threshold_cfg.is_bad(
                    self.check_value
                ):
                    self.outcome = CheckOutcome.WARN
                else:
                    self.outcome = CheckOutcome.PASS
            else:
                hint = (
                    " Is your column text based? The valid format config only works on text columns."
                    if metric_check_cfg.missing_and_valid_cfg and metric_check_cfg.missing_and_valid_cfg.valid_format
                    else ""
                )
                self.logs.error(
                    f"Cannot evaluate check: Expected a numeric value, but was {self.check_value}.{hint}",
                    location=self.check_cfg.location,
                )

    def get_cloud_diagnostics_dict(self) -> dict:
        metric_check_cfg: MetricCheckCfg = self.check_cfg
        cloud_diagnostics = {"value": self.check_value}
        # TODO Disabled these until cloud can ingest them properly
        # if self.formula_values:
        #     cloud_diagnostics["formula_values"] = self.formula_values
        # if self.failed_rows_sample_ref:
        #     cloud_diagnostics["failed_rows_sample_ref"] = self.failed_rows_sample_ref.get_cloud_diagnostics_dict()
        if metric_check_cfg.fail_threshold_cfg is not None:
            cloud_diagnostics["fail"] = metric_check_cfg.fail_threshold_cfg.to_soda_cloud_diagnostics_json()
        if metric_check_cfg.warn_threshold_cfg is not None:
            cloud_diagnostics["warn"] = metric_check_cfg.warn_threshold_cfg.to_soda_cloud_diagnostics_json()
        if self.failed_rows_sample_ref:
            cloud_diagnostics["failedRowsFile"] = self.failed_rows_sample_ref.get_cloud_diagnostics_dict()
        return cloud_diagnostics

    def get_log_diagnostic_dict(self) -> dict:
        log_diagnostics = {"check_value": self.check_value}
        if self.formula_values:
            log_diagnostics.update(self.formula_values)
        if self.failed_rows_sample_ref:
            log_diagnostics["failed_rows_sample_ref"] = str(self.failed_rows_sample_ref)
        return log_diagnostics



import logging
from dataclasses import dataclass
from enum import Enum
from textwrap import indent
from typing import Optional

from soda.common.exception_helper import get_exception_stacktrace
from soda.sodacl.location import Location

logger = logging.getLogger("soda.scan")


class LogLevel(Enum):
    WARNING = "WARNING"
    ERROR = "ERROR"
    INFO = "INFO"
    DEBUG = "DEBUG"


@dataclass
class Log:

    __log_level_mappings = {
        LogLevel.ERROR: logging.ERROR,
        LogLevel.WARNING: logging.WARNING,
        LogLevel.INFO: logging.INFO,
        LogLevel.DEBUG: logging.DEBUG,
    }

    level: LogLevel
    message: str
    location: Optional[Location]
    doc: Optional[str]
    exception: Optional[BaseException]

    def __str__(self):
        location_str = f" | {self.location}" if self.location else ""
        doc_str = f" | https://go.soda.io/{self.doc}" if self.doc else ""
        exception_str = f" | {self.exception}" if self.exception else ""
        return f"{self.level.value.ljust(7)}| {self.message}{location_str}{doc_str}{exception_str}"

    def get_cloud_dict(self) -> dict:
        log_cloud_dict = {"level": self.level.value, "message": self.message}
        if self.location:
            log_cloud_dict["location"] = self.location.to_soda_cloud_json()
        if self.doc:
            log_cloud_dict["doc"] = self.doc
        if self.exception:
            log_cloud_dict["exception"] = get_exception_stacktrace(self.exception)
        return log_cloud_dict

    @staticmethod
    def log_errors(error_logs):
        logger.info(f"ERRORS:")
        for error_log in error_logs:
            error_log.log_to_python_logging()

    def log_to_python_logging(self):
        python_log_level = Log.__log_level_mappings[self.level]
        logger.log(python_log_level, self.message)
        if self.exception is not None:
            exception_str = str(self.exception)
            exception_str = indent(text=exception_str, prefix="  | ")
            logger.log(python_log_level, exception_str)
            # Logging the stack trace
            from soda.scan import verbose

            if verbose:
                stacktrace = get_exception_stacktrace(self.exception)
                indented_stacktrace = indent(text=stacktrace, prefix="  | ")
                logger.log(python_log_level, f"  | Stacktrace:")
                logger.log(python_log_level, indented_stacktrace)
        if self.level in [LogLevel.WARNING, LogLevel.ERROR] and self.location is not None:
            logger.log(python_log_level, f"  +-> {self.location}")
        if isinstance(self.doc, str):
            logger.log(python_log_level, f"  +-> See https://go.soda.io/{self.doc}")



from typing import List, Optional

from soda.execution.aggregation_query import AggregationQuery
from soda.execution.duplicates_query import DuplicatesQuery
from soda.execution.numeric_query_metric import NumericQueryMetric
from soda.execution.partition import Partition
from soda.execution.query import Query
from soda.execution.schema_metric import SchemaMetric
from soda.execution.schema_query import SchemaQuery


class PartitionQueries:
    def __init__(self, partition: Partition):
        from soda.execution.data_source_scan import DataSourceScan

        self.logs = partition.data_source_scan.scan._logs
        self.partition: Partition = partition
        self.data_source_scan: DataSourceScan = partition.data_source_scan
        self.schema_query: Optional[SchemaQuery] = None
        self.aggregation_queries: List[AggregationQuery] = []
        self.duplicate_queries: List[DuplicatesQuery] = []

    def add_metric(self, metric: "Metric"):

        if isinstance(metric, SchemaMetric):
            self.schema_query = SchemaQuery(self.partition, metric)

        elif isinstance(metric, NumericQueryMetric):
            sql_aggregation_expression = metric.get_sql_aggregation_expression()
            if sql_aggregation_expression:
                max_aggregation_fields = self.data_source_scan.data_source.get_max_aggregation_fields()
                if (
                    len(self.aggregation_queries) == 0
                    or len(self.aggregation_queries[-1].metrics) >= max_aggregation_fields
                ):
                    aggregation_query_index = len(self.aggregation_queries)
                    aggregation_query = AggregationQuery(self.partition, aggregation_query_index)
                    self.aggregation_queries.append(aggregation_query)
                else:
                    aggregation_query = self.aggregation_queries[-1]
                aggregation_query.add_metric(sql_aggregation_expression, metric)

            elif metric.name == "duplicate_count":
                self.duplicate_queries.append(DuplicatesQuery(self.partition, metric))

            else:
                self.logs.error(f"Unsupported metric {metric.name}")
        else:
            self.logs.error(f"Unsupported metric {metric.name} ({type(metric).__name__})")

    def get_queries(self) -> List[Query]:
        queries: List[Query] = []
        if self.schema_query:
            queries.append(self.schema_query)
        queries.extend(self.aggregation_queries)
        queries.extend(self.duplicate_queries)
        return queries



from __future__ import annotations

import collections
from abc import ABC, abstractmethod

from soda.execution.check_outcome import CheckOutcome
from soda.execution.column import Column
from soda.execution.identity import ConsistentHashBuilder
from soda.execution.metric import Metric
from soda.sampler.sample_ref import SampleRef
from soda.soda_cloud.historic_descriptor import HistoricDescriptor
from soda.sodacl.check_cfg import CheckCfg
from soda.sodacl.distribution_check_cfg import DistributionCheckCfg


class Check(ABC):
    @staticmethod
    def create(
        check_cfg: CheckCfg,
        partition: Partition | None = None,
        column: Column | None = None,
        data_source_scan: DataSourceScan | None = None,
    ) -> Check:
        from soda.sodacl.anomaly_metric_check_cfg import AnomalyMetricCheckCfg
        from soda.sodacl.change_over_time_metric_check_cfg import (
            ChangeOverTimeMetricCheckCfg,
        )
        from soda.sodacl.freshness_check_cfg import FreshnessCheckCfg
        from soda.sodacl.metric_check_cfg import MetricCheckCfg
        from soda.sodacl.reference_check_cfg import ReferenceCheckCfg
        from soda.sodacl.row_count_comparison_check_cfg import (
            RowCountComparisonCheckCfg,
        )
        from soda.sodacl.schema_check_cfg import SchemaCheckCfg
        from soda.sodacl.user_defined_failed_rows_check_cfg import (
            UserDefinedFailedRowsCheckCfg,
        )
        from soda.sodacl.user_defined_failed_rows_expression_check_cfg import (
            UserDefinedFailedRowsExpressionCheckCfg,
        )

        if isinstance(check_cfg, ChangeOverTimeMetricCheckCfg):
            from soda.execution.change_over_time_metric_check import (
                ChangeOverTimeMetricCheck,
            )

            return ChangeOverTimeMetricCheck(check_cfg, data_source_scan, partition, column)

        elif isinstance(check_cfg, AnomalyMetricCheckCfg):
            from soda.execution.anomaly_metric_check import AnomalyMetricCheck

            return AnomalyMetricCheck(check_cfg, data_source_scan, partition, column)

        elif isinstance(check_cfg, MetricCheckCfg):
            from soda.execution.metric_check import MetricCheck

            return MetricCheck(check_cfg, data_source_scan, partition, column)

        elif isinstance(check_cfg, SchemaCheckCfg):
            from soda.execution.schema_check import SchemaCheck

            return SchemaCheck(check_cfg, data_source_scan, partition)
        elif isinstance(check_cfg, UserDefinedFailedRowsExpressionCheckCfg):
            from soda.execution.user_defined_failed_rows_expression_check import (
                UserDefinedFailedRowsExpressionCheck,
            )

            return UserDefinedFailedRowsExpressionCheck(check_cfg, data_source_scan, partition)
        elif isinstance(check_cfg, UserDefinedFailedRowsCheckCfg):
            from soda.execution.user_defined_failed_rows_check import (
                UserDefinedFailedRowsCheck,
            )

            return UserDefinedFailedRowsCheck(check_cfg, data_source_scan, partition)
        elif isinstance(check_cfg, RowCountComparisonCheckCfg):
            from soda.execution.row_count_comparison_check import (
                RowCountComparisonCheck,
            )

            return RowCountComparisonCheck(check_cfg, data_source_scan, partition)
        elif isinstance(check_cfg, ReferenceCheckCfg):
            from soda.execution.reference_check import ReferenceCheck

            return ReferenceCheck(check_cfg, data_source_scan, partition)
        elif isinstance(check_cfg, FreshnessCheckCfg):
            from soda.execution.freshness_check import FreshnessCheck

            return FreshnessCheck(check_cfg, data_source_scan, partition, column)

        elif isinstance(check_cfg, DistributionCheckCfg):
            from soda.execution.distribution_check import DistributionCheck

            return DistributionCheck(check_cfg, data_source_scan, partition, column)

        raise RuntimeError(f"Bug: Unsupported check type {type(check_cfg)}")

    def __init__(
        self,
        check_cfg: CheckCfg,
        data_source_scan: DataSourceScan,
        partition: Partition | None,
        column: Column | None,
        name: str | None,
    ):
        from soda.execution.partition import Partition

        self.name: str = name
        self.check_cfg: CheckCfg = check_cfg
        self.logs = data_source_scan.scan._logs
        self.data_source_scan = data_source_scan
        self.partition: Partition = partition
        self.column: Column = column
        self.metrics: dict[str, Metric] = {}
        self.historic_descriptors: dict[str, HistoricDescriptor] = {}
        self.cloud_check_type = "metricThreshold"
        # in the evaluate method, checks can optionally extract a failed rows sample ref from the metric
        self.failed_rows_sample_ref: SampleRef | None = None

        # Attribute for automated monitoring
        self.archetype = None

        # Check evaluation outcome
        self.outcome: CheckOutcome | None = None

    def create_definition(self) -> str:
        check_cfg: CheckCfg = self.check_cfg
        from soda.common.yaml_helper import to_yaml_str

        if isinstance(check_cfg.source_configurations, dict):
            return to_yaml_str({check_cfg.source_header: [{check_cfg.source_line: check_cfg.source_configurations}]})
        else:
            return f"{check_cfg.source_header}:\n  {check_cfg.source_line}"

    def create_identity(self) -> str:
        check_cfg: CheckCfg = self.check_cfg
        from soda.common.yaml_helper import to_yaml_str

        if isinstance(check_cfg.source_configurations, dict):
            identity = check_cfg.source_configurations.get("identity")
            if isinstance(identity, str):
                return identity

        hash_builder = ConsistentHashBuilder()
        # Note: In case of for each table, the check_cfg.source_header will contain the actual table name as well
        hash_builder.add(check_cfg.source_header)
        hash_builder.add(check_cfg.source_line)

        # If the check is automated monitoring append "automated_monitoring" tag
        if hasattr(check_cfg, "is_automated_monitoring") and check_cfg.is_automated_monitoring:
            hash_builder.add("automated_monitoring")

        if isinstance(check_cfg.source_configurations, dict):

            identity_source_configurations = dict(check_cfg.source_configurations)
            # The next lines ensures that configuration properties 'name' and 'identity' are ignored
            # for computing the check identity
            identity_source_configurations.pop("name", None)
            if len(identity_source_configurations) > 0:
                # The next line ensures that ordering of the check configurations don't matter for identity
                identity_source_configurations = collections.OrderedDict(sorted(identity_source_configurations.items()))
                identity_source_configurations_yaml = to_yaml_str(identity_source_configurations)
                hash_builder.add(identity_source_configurations_yaml)
        return hash_builder.get_hash()

    @staticmethod
    def __check_source_to_yaml(source_header: str, source_line: str, source_configurations: dict | None) -> str:
        from soda.common.yaml_helper import to_yaml_str

        if not isinstance(source_configurations, dict):
            return f"{source_header}:\n  {source_line}"
        return to_yaml_str({source_header: [{source_line: source_configurations}]})

    def get_cloud_dict(self):
        from soda.execution.column import Column
        from soda.execution.partition import Partition

        cloud_dict = {
            "identity": self.create_identity(),
            "name": self.generate_soda_cloud_check_name(),
            "type": self.cloud_check_type,
            "definition": self.create_definition(),
            "location": self.check_cfg.location.to_soda_cloud_json(),
            "dataSource": self.data_source_scan.data_source.data_source_name,
            "table": Partition.get_table_name(self.partition),
            # "filter": Partition.get_partition_name(self.partition), TODO: re-enable once backend supports the property.
            "column": Column.get_partition_name(self.column),
            "metrics": [metric.identity for metric in self.metrics.values()],
            "outcome": self.outcome.value if self.outcome else None,
            "diagnostics": self.get_cloud_diagnostics_dict(),
        }
        # Update dict if automated monitoring is running
        if self.archetype is not None:
            cloud_dict.update({"archetype": self.archetype})

        return cloud_dict

    def generate_soda_cloud_check_name(self) -> str:
        if self.check_cfg.name:
            return self.check_cfg.name

        name = self.check_cfg.source_line

        if self.check_cfg.source_configurations:
            source_cfg = dict(self.check_cfg.source_configurations)

            if source_cfg.get("warn"):
                name += f" warn {source_cfg['warn']}"

            if source_cfg.get("fail"):
                name += f" fail {source_cfg['fail']}"

        return name

    @abstractmethod
    def get_cloud_diagnostics_dict(self) -> dict:
        pass

    def evaluate(self, metrics: dict[str, Metric], historic_values: dict[str, object]):
        raise NotImplementedError("Implement this abstract method")

    def get_log_diagnostic_lines(self) -> list[str]:
        log_diagnostic_lines = []

        log_diagnostic_dict = self.get_log_diagnostic_dict()
        if log_diagnostic_dict:
            for key, value in log_diagnostic_dict.items():
                log_diagnostic_lines.append(f"{key}: {value}")

        return log_diagnostic_lines

    def get_log_diagnostic_dict(self) -> dict:
        return self.get_cloud_diagnostics_dict()

    def get_summary(self) -> str:
        return self.check_cfg.name if self.check_cfg.name else self.check_cfg.source_line

    def __str__(self):
        log_diagnostic_lines = self.get_log_diagnostic_lines()
        if log_diagnostic_lines:
            diagnostics_text = ", ".join(log_diagnostic_lines)
            return f"[{self.check_cfg.source_line}] {self.outcome} ({diagnostics_text})"
        else:
            return f"[{self.check_cfg.source_line}] {self.outcome}"



from dataclasses import dataclass
from typing import List


@dataclass
class SampleColumn:
    name: str
    type: str

    def get_cloud_dict(self):
        return {"name": self.name, "type": self.type}


@dataclass
class SampleSchema:
    columns: List[SampleColumn]



from typing import TYPE_CHECKING, Dict, List

from soda.execution.data_source import DataSource
from soda.execution.discover_tables_run import DiscoverTablesRun
from soda.execution.metric import Metric
from soda.execution.profile_columns_run import ProfileColumnsRun
from soda.execution.query import Query
from soda.execution.table import Table
from soda.sodacl.data_source_scan_cfg import DataSourceScanCfg

if TYPE_CHECKING:
    from soda.scan import Scan


class DataSourceScan:
    def __init__(
        self,
        scan: "Scan",
        data_source_scan_cfg: DataSourceScanCfg,
        data_source: DataSource,
    ):
        from soda.execution.metric import Metric
        from soda.execution.table import Table

        self.scan: Scan = scan
        self.data_source_scan_cfg: DataSourceScanCfg = data_source_scan_cfg
        self.metrics: List[Metric] = []
        self.data_source: DataSource = data_source
        self.tables: Dict[str, Table] = {}
        self.queries: List[Query] = []

    def get_or_create_table(self, table_name: str) -> Table:
        table = self.tables.get(table_name)
        if table is None:
            table = Table(self, table_name)
            self.tables[table_name] = table
        return table

    def resolve_metric(self, metric: "Metric") -> Metric:
        """
        If the metric is not added before, this method will:
         - Add the metric to scan.metrics
         - Ensure the metric is added to the appropriate query (if applicable)
        """
        existing_metric = self.scan._find_existing_metric(metric)
        if existing_metric:
            existing_metric.merge_checks(metric)
            return existing_metric
        self.scan._add_metric(metric)
        metric.ensure_query()
        return metric

    def get_queries(self):
        return

    def execute_queries(self):
        all_data_source_queries: List[Query] = []
        for table in self.tables.values():
            for partition in table.partitions.values():
                partition_queries = partition.collect_queries()
                all_data_source_queries.extend(partition_queries)
        all_data_source_queries.extend(self.queries)

        for query in all_data_source_queries:
            query.execute()

    def create_automated_monitor_run(self, automated_monitoring_cfg, scan):
        from soda.execution.automated_monitoring_run import AutomatedMonitoringRun

        return AutomatedMonitoringRun(self, automated_monitoring_cfg)

    def create_profile_columns_run(self, profile_columns_cfg, scan):
        return ProfileColumnsRun(self, profile_columns_cfg)

    def create_discover_tables_run(self, discover_tables_cfg, scan):
        return DiscoverTablesRun(self, discover_tables_cfg)



import logging
from typing import List

from ruamel.yaml import YAML
from ruamel.yaml.error import MarkedYAMLError
from soda.common.logs import Logs
from soda.sodacl.location import Location

logger = logging.getLogger(__name__)


class PathElement:
    def __init__(self, location: Location, value=None):
        self.location: Location = location
        self.value = value


class PathStack:
    def __init__(self, file_path: str):
        self.file_path = file_path
        self.path_elements: List[PathElement] = [PathElement(location=Location(file_path))]

    def set_antlr_collection_in_file_path_element(self, file_dict: dict):
        self.path_elements[0].value = file_dict

    def push_path_element(self, key, value) -> Location:
        location = self.create_location(key)
        self.path_elements.append(PathElement(location=location, value=value))
        return location

    def pop_path_element(self):
        self.path_elements.pop()
        return self.path_elements[-1].location

    def create_location(self, key):
        last_path_element = self.path_elements[-1]
        collection = last_path_element.value
        if isinstance(collection, dict) and isinstance(key, str):
            item_data = collection.lc.data.get(key)
        elif isinstance(collection, list) and isinstance(key, int):
            item_data = collection.lc.data[key]
        else:
            return last_path_element.location
        line = item_data[0] + 1
        col = item_data[1] + 1
        # path_str should only be the nested keys inside the yaml, not the file_path
        # first path element is the file path so we skip that in the path_str
        return Location(self.file_path, line, col)

    def get_value(self, key):
        return self.path_elements[-1].value.get(key)


class Parser:
    def __init__(self, file_path: str, logs: Logs):
        self.path_stack: PathStack = PathStack(file_path)
        self.location = self.path_stack.path_elements[0].location
        self.logs = logs

    def _parse_yaml_str(self, yaml_str: str) -> dict:
        try:
            yaml = YAML()
            yaml.preserve_quotes = True
            file_dict = yaml.load(yaml_str)
            self.path_stack.set_antlr_collection_in_file_path_element(file_dict)
            return file_dict
        except MarkedYAMLError as e:
            location = self._get_location_from_yaml_error(e)
            self.logs.error(f"YAML syntax error", location=location, exception=e)

    def _get_location_from_yaml_error(self, e):
        mark = e.context_mark if e.context_mark else e.problem_mark
        return Location(
            file_path=self.path_stack.file_path,
            line=mark.line + 1,
            col=mark.column + 1,
        )

    def _push_path_element(self, key, value):
        self.location = self.path_stack.push_path_element(key, value)

    def _pop_path_element(self):
        self.location = self.path_stack.pop_path_element()

    def _get_required(self, key: str, value_type: type):
        return self.__get_value(key, required=True, value_type=value_type)

    def _get_optional(self, key: str, value_type: type):
        return self.__get_value(key, required=False, value_type=value_type)

    def __get_value(self, key: str, required: bool, value_type: type):
        value = self.path_stack.get_value(key)
        if value is None:
            if required:
                self.logs.error(f"{key} is required", location=self.location)
        elif not isinstance(value, value_type):
            self.logs.error(
                f'Invalid YAML structure: key "{key}" must contain a {value_type.__name__}, but was {type(value).__name__}',
                location=self.location,
            )
            return None
        return value

    def _resolve_jinja(self, value: str, variables: dict = None):
        from soda.common.jinja import Jinja

        """
        Resolves the given string with Jinja templating and then parses the resolved string as a configuration yaml
        """
        if isinstance(value, str) and "$" in value:
            try:
                return Jinja.resolve(template=value, variables=variables)
            except BaseException as e:
                self.logs.error(
                    f"Could not resolve {value} with Jinja: {e}",
                    e,
                    location=self.location,
                )
        return value



from typing import Optional


class MetricCfg:
    def __init__(self):
        self.metric_name: str = None
        self.table_name: Optional[str] = None
        self.column_name: Optional[str] = None
        self.title: str = None



#!/usr/bin/env python
import sys

from setuptools import find_namespace_packages, setup

if sys.version_info < (3, 8):
    print("Error: Soda Core requires at least Python 3.8")
    print("Error: Please upgrade your Python version to 3.8 or later")
    sys.exit(1)

package_name = "soda-core"
# Managed by tbump - do not change manually
package_version = "3.0.0b13"
description = "Soda Core"

# long_description = (pathlib.Path(__file__).parent.parent / "README.md").read_text()

requires = [
    "markupsafe",
    "Jinja2~=3.0",
    "click",
    "ruamel.yaml",
    "requests",
    "antlr4-python3-runtime~=4.9.3",
    "opentelemetry-exporter-otlp-proto-http",
]

setup(
    name=package_name,
    version=package_version,
    author="Soda Data N.V.",
    author_email="info@soda.io",
    description="Soda Core library & CLI",
    long_description="long_description",  # TODO: fix this as it fails in tox
    long_description_content_type="text/markdown",
    packages=find_namespace_packages(include=["soda*"]),
    install_requires=requires,
    entry_points={"console_scripts": ["soda=soda.cli.cli:main"]},
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    python_requires=">=3.8",
)



from __future__ import annotations


class DataType:
    """
    Generic data types that data_sources can translate to specific types
    """

    TEXT = "text"
    INTEGER = "integer"
    DECIMAL = "decimal"
    DATE = "date"
    TIME = "time"
    TIMESTAMP = "timestamp"
    TIMESTAMP_TZ = "timestamptz"
    BOOLEAN = "boolean"



from soda.execution.query import Query


class SampleQuery(Query):
    def __init__(self, data_source_scan: "DataSourceScan", metric: "Metric", sample_type: str, sql: str):
        """
        :param sample_type: Eg 'failed_rows'
        """
        super().__init__(
            data_source_scan=data_source_scan,
            partition=metric.partition,
            column=metric.column,
            unqualified_query_name=f"{sample_type}[{metric.name}]",
            sql=sql,
        )
        self.metric = metric

    def execute(self):
        self.store()
        self.metric.failed_rows_sample_ref = self.sample_ref
